"""
Jsonification and unjsonification for standard Python types.

* datetime.datetime
* datetime.date
* datetime.time
* datetime.timezone
* decimal.Decimal
* pathlib.Path
* zoneinfo.ZoneInfo
* enum.Enum
* ranges

"""

import dataclasses
import decimal
import enum
import pathlib
import re
import zoneinfo

from types import NoneType

from jsno.jsonify import jsonify
from jsno.unjsonify import unjsonify, typecheck, raise_error, cast


# marking types to be jsonified as strings


def jsonify_to_string(value):
    return str(value)


def unjsonify_from_string(value, as_type):
    typecheck(value, str, as_type)
    return cast(value, as_type)


def jsonify_as_string(type_, exceptions=()):
    jsonify.register(type_)(jsonify_to_string)

    if exceptions:

        @unjsonify.register(type_)
        def _(value, as_type):
            typecheck(value, str, as_type)
            try:
                return cast(value, as_type)
            except Exception as exc:
                if not isinstance(exc, exceptions):
                    raise

                detail = exc.args[0]

            raise_error(value, as_type, detail)

    else:
        unjsonify.register(type_)(unjsonify_from_string)

    return type_


# NoneType, jsonify is not needed

@unjsonify.register(NoneType)
def _(value, as_type):
    typecheck(value, NoneType, as_type)
    return None


# str

jsonify_as_string(str)

# bool, jsonify is not needed


@unjsonify.register(bool)
def _(value, as_type):
    typecheck(value, bool, as_type)
    return cast(value, as_type)


# int


@jsonify.register(int)
def _(value):
    return int(value)


@unjsonify.register(int)
def _(value, as_type):
    typecheck(value, int, as_type)
    return cast(value, as_type)


# float


@jsonify.register(float)
def _(value):
    return float(value)


@unjsonify.register(float)
def _(value, as_type):
    typecheck(value, (float, int), as_type)
    return cast(value, as_type)


# enums


@jsonify.register(enum.Enum)
def _(enum):
    return enum.name


@unjsonify.register(enum.Enum)
def _(value, as_type):
    typecheck(value, str, as_type)
    try:
        return getattr(as_type, value)
    except AttributeError:
        pass

    raise_error(value, as_type)


# zoneinfo

jsonify_as_string(zoneinfo.ZoneInfo, exceptions=(zoneinfo.ZoneInfoNotFoundError))

# decimal


@jsonify.register(decimal.Decimal)
def _(value):
    """
    Always jsnonify Decimal as a string.
    """
    return str(value)


@unjsonify.register(decimal.Decimal)
def _(value, as_type):
    typecheck(value, (str, int), as_type)

    return as_type(value)


# pathlib.Path

jsonify_as_string(pathlib.Path)

# complex numbers

jsonify_as_string(complex)


@jsonify.register(range)
def _(value):
    result = {"start": value.start, "stop": value.stop}
    if value.step != 1:
        result["step"] = value.step
    return result


@dataclasses.dataclass(slots=True, frozen=True)
class Range:
    start: int
    stop: int
    step: int | None = None


@unjsonify.register(range)
def _(value, as_type):
    it = unjsonify[Range](value)
    if it.step == 0:
        raise_error(value, as_type, "Range step must not be zero")

    return as_type(it.start, it.stop, it.step or 1)


# re.Pattern


@jsonify.register(re.Pattern)
def _(value):
    return value.pattern


@unjsonify.register(re.Pattern)
def _(value, as_type):
    return re.compile(value)



from datetime import datetime
from typing import TypedDict, NotRequired, Required, NamedTuple

import pytest

from jsno import jsonify, unjsonify, UnjsonifyError


class ApiKey(TypedDict):
    name: NotRequired[str]
    value: str
    created_at: datetime


def test_unjsonify_typed_dict():
    apikey = unjsonify[ApiKey]({"value": "XXX", "created_at": "2023-08-05T07:22:33"})
    assert apikey == {"value": "XXX", "created_at": datetime(2023, 8, 5, 7, 22, 33)}


class ApiKey2(TypedDict, total=False):
    name: str
    value: Required[str]
    created_at: datetime


def test_unjsonify_nontotal_typed_dict():
    apikey = unjsonify[ApiKey2]({"value": "XXX", "created_at": "2023-08-05T07:22:33"})
    assert apikey == {"value": "XXX", "created_at": datetime(2023, 8, 5, 7, 22, 33)}


def test_unjsonify_nontotal_typed_dict_missing_required():
    with pytest.raises(UnjsonifyError):
        unjsonify[ApiKey2]({"name": "key", "created_at": "2023-08-05T07:22:33"})


def test_unjsonify_nontotal_typed_dict_extra_key():
    with pytest.raises(UnjsonifyError):
        unjsonify[ApiKey2]({"value": "XXX", "treated_at": "2023-08-05T07:22:33"})


def test_unjsonify_nontotal_typed_dict_ignore_extra_key():
    with unjsonify.ignore_extra_keys():
        apikey = unjsonify[ApiKey2]({"value": "XXX", "treated_at": "2023-08-05T07:22:33"})
        assert apikey == {"value": "XXX"}


class LogEntry(NamedTuple):
    date: datetime
    message: str


entry = LogEntry(date=datetime(2023, 8, 5, 8, 32, 10), message="fail")
json = ["2023-08-05T08:32:10", "fail"]


def test_jsnoify_namedtuple():
    assert jsonify(entry) == json


def test_unjsonify_namedtuple():
    assert unjsonify[LogEntry](json) == entry


def test_unjsonify_namedtuple_error():
    with pytest.raises(UnjsonifyError):
        unjsonify[LogEntry](["2023-08-05T08:32:10"])



import dataclasses
import functools
import types

from collections.abc import Mapping
from typing import Annotated, Any, Union, Literal, NewType, get_args, get_origin

from jsno.extra_data import get_extra_data_configuration, Ignore
from jsno.property_name import resolve_field_name
from jsno.utils import contextvar, DictWithoutKey
from jsno.variant import get_variantfamily


unjsonify_context = contextvar(on_extra_key="error")


class UnjsonifyError(TypeError):
    pass


def raise_error(value: Any, as_type: Any, detail=None):
    raise UnjsonifyError(f"Cannot unjsonify as {as_type}", value, detail)


def typecheck(value: Any, jsontype: type, as_type: Any) -> None:
    """
    Check if the value is an instance of the type given as `jsontype`
    and if not, raise an appropriate UnjsonifyError
    """

    if not isinstance(value, jsontype):
        raise_error(value, as_type)


def cast(value: Any, as_type: Any) -> Any:
    """
    Make sure that the value given has given type, and if not,
    try casting it to that type.
    """

    if isinstance(as_type, type):
        origin_type = as_type
    else:
        origin_type = get_origin(as_type)

    if isinstance(value, origin_type):
        return value

    try:
        return origin_type(value)
    except ValueError as exc:
        detail = exc.args[0]

    raise_error(value, as_type, detail)


@functools.singledispatch
def unjsonify_type(as_type):
    if dataclasses.is_dataclass(as_type):
        return get_unjsonify_dataclass(as_type)

    raise TypeError(f"Unjsonify not defined for {as_type.__qualname__}")


def handle_extra_keys(value, result, as_type):
    extra_data_property = get_extra_data_configuration(as_type)
    if extra_data_property is not None:
        if extra_data_property is Ignore:
            return

        result[extra_data_property] = {
            key: value[key] for key in value if key not in result
        }

    elif unjsonify_context.on_extra_key == "error":
        extra_keys = {key for key in value if key not in result}
        raise UnjsonifyError(
            f"Extra keys for {as_type.__qualname__}: {', '.join(extra_keys)}"
        )


def get_unjsonify_dataclass(as_type):

    fields = [
        (field.name, json_name, unjsonify[field.type])
        for field in dataclasses.fields(as_type)
        if (json_name := resolve_field_name(field))
    ]

    def specialized(value):
        typecheck(value, (dict, Mapping), as_type)

        # collect all properties in the input value that match any of
        # the dataclass fields
        kwargs = {
            name: unjsonify_(value.get(json_name))
            for (name, json_name, unjsonify_) in fields
            if json_name in value
        }
        if len(kwargs) < len(value):
            handle_extra_keys(value, kwargs, as_type)

        try:
            return as_type(**kwargs)
        except TypeError as exc:
            detail = exc.args[0]

        raise_error(value, as_type, detail)

    return specialized


def get_unjsonify_variant(as_type, family):

    cache = {}

    def specialized(value):
        typecheck(value, (dict, Mapping), as_type)

        label_name = family.label_name

        label = value.get(label_name)
        if label is None:
            raise_error(value, as_type, f"missing {label}")

        func = cache.get(label)
        if func is None:
            variant_type = family.get_variant(label)
            if variant_type is None or not issubclass(variant_type, as_type):
                raise_error(value, as_type, f"unknown {label_name} label: {label}")

            func = unjsonify._dispatch(variant_type)
            cache[label] = func

        return func(DictWithoutKey(base=value, key=label_name))

    return specialized


def get_unjsonify_literal(as_type):
    options = get_args(as_type)

    return lambda value: value if value in options else raise_error(value, as_type)


class Unjsonify:
    def __init__(self):
        self._cache = {}

    def _dispatch(self, type_):
        if isinstance(type_, NewType):
            type_ = type_.__supertype__

        origin = get_origin(type_)
        if origin:
            # special cases needed for constructs in typing module, as
            # singledispatch fails cannot handle Union and Literal
            if origin is Union:
                return get_unjsonify_union(type_)
            elif origin is Literal:
                return get_unjsonify_literal(type_)

        # covers list[X], dict[K,V], etc.
        try:
            func = unjsonify_type.dispatch(origin or type_)
        except TypeError:
            raise UnjsonifyError(f"Cannot unjsonify {type_}")

        return func(type_)

    def __getitem__(self, type_):
        unjsonify = self._cache.get(type_)
        if unjsonify is None:
            if isinstance(type_, type) and (family := get_variantfamily(type_)):
                unjsonify = get_unjsonify_variant(type_, family)
            else:
                unjsonify = self._dispatch(type_)
            self._cache[type_] = unjsonify
        return unjsonify

    def register(self, type_):
        def decorator(func):
            @self.register_factory(type_)
            def _(as_type):
                return lambda value: func(value, as_type)

            return func

        return decorator

    def register_factory(self, type_):
        self._cache.clear()
        return unjsonify_type.register(type_)

    def context(self, **kwargs):
        return unjsonify_context(**kwargs)

    def ignore_extra_keys(self):
        return self.context(on_extra_key="ignore")


unjsonify = Unjsonify()


@unjsonify.register_factory(types.UnionType)
def get_unjsonify_union(as_type):
    """
    Unjsonify a Union type. Tries each option at a time, and
    selects the first one that matches.
    """

    options = get_args(as_type)

    def specialized(value):
        for type_option in options:
            try:
                return unjsonify[type_option](value)
            except UnjsonifyError:
                # try next option
                continue

        raise UnjsonifyError(f"Cannot unjsonify as {as_type}", value)

    return specialized


@unjsonify.register(Any)
def _(value, as_type):
    """Unjsonify Any type: just return the value"""
    return value


@functools.singledispatch
def validate_annotation(arg, value):
    pass


@unjsonify.register(Annotated)
def _(value, as_type):
    args = get_args(as_type)
    result = unjsonify[args[0]](value)

    ix = 1
    while ix < len(args):
        try:
            validate_annotation(args[ix], result)
            ix += 1
            continue
        except ValueError as exc:
            detail = exc.args[0]

        raise_error(result, as_type, detail)

    return result



from dataclasses import dataclass
import typing

from jsno.utils import Annotation


@dataclass(frozen=True, slots=True)
class PropertyName(Annotation):
    name: str


def property_name(name):
    return PropertyName(name)


def resolve_field_name(field):
    if typing.get_origin(field.type) is typing.Annotated:
        for arg in typing.get_args(field.type):
            if isinstance(arg, PropertyName):
                return arg.name

    return field.name



import functools
import inspect
import typing

from jsno.unjsonify import UnjsonifyError, unjsonify, handle_extra_keys, typecheck


def unwrap_type(type_):
    """
    Unwrap the type for a typed dict key that might have been
    annotated with `Required` or `NotRequired`
    """

    origin = typing.get_origin(type_)
    if origin is typing.NotRequired or origin is typing.Required:
        return typing.get_args(type_)[0]
    else:
        return type_


@functools.cache
def unjsonify_typeddict_factory(as_type):

    annotations = [
        (key, unjsonify[unwrap_type(type_)], key in as_type.__required_keys__)
        for (key, type_) in inspect.get_annotations(as_type).items()
    ]

    def specialized(value):
        typecheck(value, (dict, typing.Mapping), as_type)

        result = {}

        for (key, unjsonify_, required) in annotations:
            if key in value:
                result[key] = unjsonify_(value[key])
            elif required:
                raise UnjsonifyError(f"Required key missing in {as_type}: {key}")

        # check if all values were handled
        if len(result) < len(value):
            handle_extra_keys(value, result, as_type)

        return result

    return specialized



import re

from dataclasses import dataclass
from typing import Any, Callable

from jsno.unjsonify import validate_annotation
from jsno.utils import Annotation


class Constraint(Annotation):
    """
    Base class for consraints
    """

    def __new__(cls, *args, **kwargs):
        if cls is Constraint:
            # Creating a plain Constraint actually creates a
            # FunctionConstraint
            cls = FunctionConstraint

        obj = object.__new__(cls)
        obj.__init__(*args, **kwargs)
        return obj

    def __or__(self, that: "Constraint") -> "Constraint":
        """
        Combining constraints with or-operator
        """
        return OrConstraint(self, that)

    @staticmethod
    def regex(regex, name=None):
        return RegExConstraint(
            name=(name or f"Regular expression {regex}"),
            regex=re.compile(regex),
        )

    def evaluate(self, value) -> bool:
        raise NotImplementedError()  # pragma: no cover


@dataclass(slots=True, frozen=True)
class OrConstraint(Constraint):
    left: Constraint
    right: Constraint

    def evaluate(self, value) -> bool:
        return self.left.evaluate(value) or self.right.evaluate(value)

    @property
    def name(self):
        return f'{self.left.name} or {self.right.name}'


@dataclass(slots=True, frozen=True)
class RangeConstraint(Constraint):
    value_name = "Value"

    min: Any | None = None
    max: Any | None = None

    def evaluate(self, value) -> bool:
        return (
            (self.min is None or value >= self.min) and
            (self.max is None or value <= self.max)
        )

    @property
    def name(self):
        if self.min is None:
            return f"{self.value_name} must be at least {self.min}"
        elif self.max is None:
            return f"{self.value_name} must be at most {self.max}"
        else:
            return f"{self.value_name} must be in range {self.min}..{self.max}"


class LenConstraint(RangeConstraint):
    value_name = "Length"

    def evaluate(self, value) -> bool:
        return super().evaluate(len(value))


@dataclass(slots=True, frozen=True)
class RegExConstraint(Constraint):
    name: str
    regex: re.Pattern

    def evaluate(self, value) -> bool:
        return self.regex.fullmatch(str(value)) is not None


Constraint.range = RangeConstraint  # type: ignore
Constraint.len = LenConstraint  # type: ignore


@dataclass(slots=True, frozen=True)
class FunctionConstraint(Constraint):
    function: Callable[[Any], bool]
    name: str | None = None

    def evaluate(self, value) -> bool:
        return self.function(value)


@validate_annotation.register(Constraint)
def _(constraint, value):
    if not constraint.evaluate(value):
        if constraint.name:
            message = f"Violates constraint: {constraint.name}"
        else:
            message = "Violates a constraint"

        raise ValueError(message)



"""
Decorator for marking a class so that it will be jsonified by a method
call: obj.jsonify()
"""

from jsno.jsonify import jsonify
from jsno.unjsonify import unjsonify


def jsonify_with_method_call(value):
    return value.jsonify()


def unjsonify_with_method_call(value, as_type):
    return as_type.unjsonify(value)


def jsonify_with_method(cls: type) -> type:
    """
    Decorator for classes that registers jsonify and unjsonify
    to call the corresponding instance and class methods.
    """

    jsonify.register(cls)(jsonify_with_method_call)
    unjsonify.register(cls)(unjsonify_with_method_call)

    return cls
